"""
Purpose: To parse and validate all Server Responses
"""
import warnings

import requests as rq

from .exceptions import ServiceError


class ValidateResponse:
    """Raise custom errors"""
    def __init__(self, resp: rq.Response, show_warn: bool = True):
        """
        Validate the response here
        :param resp: Response from the server
        :param show_warn: Whether to show warning or not
        """
        self.resp = resp
        self.show_warn = show_warn
        self.raise_error()

    def raise_error(self):
        """Check for respone `status_code` to differentiate success and everything else"""
        if self.resp.status_code in (200, 202):
            return
        elif self.resp.status_code in range(200, 206):
            self.show_warnings()
            return
        raise ServiceError(dict_resp=self.resp.json())

    def show_warnings(self):
        """Show Warnings based on `M(m)essage` from the request"""
        if not self.show_warn:
            return
        _msg_ = ''
        if self.resp.json().get('Message', ''):
            _msg_ = self.resp.json()['Message']
        else:
            _msg_ = self.resp.json().get('message', '')

        if _msg_:
            warnings.warn(_msg_)



"""
Any Request or Response of a transaction must take place here
"""
from urllib import parse as urlparse
import os
import typing as ty
from typing import BinaryIO
import time
import warnings

import requests as rq

from .FileOperations import PrepareInput
from .config import HOST, JobStatus
from .parsers import ValidateResponse
from .common import ConvertTo
from .exceptions import ClientFileSizeError


class ExtractTable:
    from .__version__ import __version__
    _OUTPUT_FORMATS: set = ConvertTo.FORMATS
    _DEFAULT: str = ConvertTo.DEFAULT
    _WARNINGS: bool = True
    _WAIT_UNTIL_OUTPUT: bool = True
    VERSION = f"ExtractTable_{__version__}"

    def __init__(self, api_key: str):
        """
        Starts by creating a session
        :param api_key: API Key recieved from https://extracttable.com
        """
        self.api_key = api_key
        self._session = rq.Session()
        self._session.headers['x-api-key'] = self.api_key

        # Helpul if the user wants to dig into the actual server response
        self.ServerResponse = rq.Response

    def _make_request(self, method, host: urlparse, params: dict = None, data: dict = None, **kwargs) -> dict:
        """
        Create a server request and parse the response for validation
        :param method: Request method
        :param host: endpoint to send the request
        :param params: query params for the request
        :param data: form data for the requests
        :param kwargs: Any other that a server accepts
        :return: json response of the server
        """
        tmp = self.__dict__.copy()
        for _type, _obj in tmp.items():
            if _type not in ("api_key", "_session", "input_filename"):
                self.__delattr__(_type)

        host = host if not host.startswith("http") else host.split("/")[2]
        url = urlparse.urlunparse(('https', host, '', '', '', ''))
        self.ServerResponse = self._session.request(method, url, params=params, data=data, **kwargs)
        ValidateResponse(resp=self.ServerResponse, show_warn=self._WARNINGS)
        self.server_response = self.ServerResponse.json()
        return self.ServerResponse.json()

    def check_usage(self) -> dict:
        """
        Check the usage of the API Key is valid
        :return the plan usage of the API Key, if valid
        """
        resp = self._make_request('get', HOST.VALIDATOR)

        return resp['usage']

    def get_result(self, job_id: str, wait_time: int = 10, max_wait_time: int = 300) -> dict:
        """
        Retrieve the tabular data of a triggered job based on the JobId
        :param job_id: JobId received from an already triggered process
        :param wait_time: Time to wait before making another request
        :param max_wait_time: Maximum Time to wait before returning to the client
        :return: Tabular JSON when processed successful else helpful user info
        """
        params = {'JobId': job_id}
        resp = self._make_request('get', HOST.RESULT, params=params)
        # Loop to retrieve the output until max_wait_time is reached
        max_wait_time = int(max_wait_time)
        while self._WAIT_UNTIL_OUTPUT and resp["JobStatus"] == JobStatus.PROCESSING and max_wait_time > 0:
            time.sleep(max(10, int(wait_time)))
            max_wait_time -= wait_time
            resp = self._make_request('get', HOST.RESULT, params=params)

        return resp

    def trigger_process(self, fp: BinaryIO, dup_check: bool = False, **kwargs) -> dict:
        """
        Trigger the document to the server for processing
        :param fp: Binary file data of the input file
        :param dup_check: helps to handle idempotent requests
        :param kwargs: anyother form-data to be sent to the server
        :return: Tabular JSON when processed successful else helpful user info
        """
        max_wait_time = kwargs.pop('max_wait_time', 300)
        data = {'dup_check': dup_check, "library": kwargs.pop("library", self.VERSION)}
        data.update(kwargs)
        if "signed_filename" in data:
            resp = self._make_request('post', HOST.TRIGGER, data=data)
        else:
            resp = self._make_request('post', HOST.TRIGGER, data=data, files={'input': fp})

        # GetResult if JobId is present in the response
        # Usually happens when processing PDF files or idempotent requests
        if 'JobId' in resp and resp.get("JobStatus", "") == JobStatus.PROCESSING:
            if max_wait_time > 0:
                print("[Info]: Waiting to retrieve the output; JobId:", resp['JobId'])
            else:
                print("[Info]: JobId:", resp['JobId'])
            resp = self.get_result(resp['JobId'], max_wait_time=max_wait_time)

        return resp

    def bigfile_upload(self, filepath):
        """
        To aid big file processing by uploading the file first and triggering the process next
        :param filepath: filepath
        :return: a signed URL to upload the file
        """
        resp = self._make_request('post', HOST.BIGFILE, data={"filename": filepath})

        return resp

    def process_file(
            self,
            filepath: ty.Union[str, bytes, os.PathLike],
            pages: ty.Union[str] = "1",
            output_format: str = "df",
            dup_check: bool = False,
            indexing: bool = False,
            **kwargs
    ) -> list:
        """
        Trigge the file for processing and returns the tabular data in the user requested output format
        :param filepath: Location of the file
        :param pages : str, optional (default: '1')
                Comma-separated page numbers.
                Example: '1,3,4' or '1,4-end' or 'all'.
        :param output_format: datafram as default; Check `ExtractTable._OUTPUT_FORMATS` to see available options
        :param dup_check: Idempotent requests handler
        :param indexing: Whether to output row & column indices in the outputs other than df
        :param kwargs:
            max_wait_time: int, optional (default: 300);
                Maximum Time to wait before returning to the client
            any other form-data to be sent to the server for future considerations
        :return: user requested output in list;
        """
        # Raise a warning if unknown format is requested
        if output_format not in self._OUTPUT_FORMATS:
            warn_msg = f"Found: '{output_format}' as output_format; Allowed formats are {self._OUTPUT_FORMATS}. " \
                       f"Assigned to default format: {self._DEFAULT}"
            warnings.warn(warn_msg)

        # To use the reference when saving the output
        self.__setattr__('input_filename', os.path.basename(filepath))

        try:
            with PrepareInput(filepath, pages=pages) as infile:
                with open(infile.filepath, 'rb') as fp:
                    trigger_resp = self.trigger_process(fp, dup_check=dup_check, **kwargs)
        except ClientFileSizeError:
            big_gen = self.bigfile_upload(filepath=os.path.basename(filepath))
            with open(filepath, 'rb') as ifile:
                rq.post(big_gen['url'], data=big_gen['fields'], files={'file': ifile})
            trigger_resp = self.trigger_process(None, signed_filename=big_gen["fields"]["key"], dup_check=dup_check, **kwargs)

        for _type, _obj in trigger_resp.items():
            self.__setattr__(_type, _obj)

        result = ConvertTo(server_response=trigger_resp, output_format=output_format, indexing=indexing).output
        return result

    def save_output(self, output_folder: os.PathLike = "", output_format: str = "csv", indexing: bool = False):
        """
        Save the objects of session data to user preferred location or a default folder
        :param output_folder: user preferred output location; default tmp directory
        :param output_format: needed only for tables CSV or XLSX
        :param indexing: row & column index consideration in the output
        :return: location of the output
        """
        input_fname = self.input_filename.rsplit('.')[0]

        output_format = output_format.lower()
        if output_format not in ("csv", "xlsx"):
            output_format = "csv"
            warnings.warn("Invalid 'output_format' given. Defaulted to 'csv'")

        table_outputs_path = ConvertTo(server_response=self.server_response, output_format=output_format, indexing=indexing).output

        if output_folder:
            if not os.path.exists(output_folder):
                output_folder = os.path.split(table_outputs_path[0])[0]
                warnings.warn(f"Your output_folder not exists. Saving the outputs to {output_folder}")
            else:
                for each_tbl_path in table_outputs_path:
                    os.replace(each_tbl_path, os.path.join(output_folder, input_fname+os.path.basename(each_tbl_path)))

        else:
            output_folder = os.path.split(table_outputs_path[0])[0]

        for each_page in self.server_response.get("Lines", []):
            page_txt_fname = os.path.join(output_folder, f"{input_fname}_Page_{str(each_page['Page'])}.txt")
            page_txt = [each_line['Line'] for each_line in each_page['LinesArray']]
            with open(page_txt_fname, "w", encoding="utf-8") as ofile:
                ofile.write("\n".join(page_txt))

        return output_folder



"""
Preprocess the output received from server and interface as a final result to the client
"""
import os
import re
import tempfile
import warnings
import collections
from statistics import mode
from typing import List

import pandas as pd


class ConvertTo:
    FORMATS = {"df", "dataframe", "json", "csv", "dict", "xlsx", "excel"}
    DEFAULT = "df"

    def __init__(self, server_response: dict, output_format: str = DEFAULT, indexing: bool = False, table_obj="TableJson"):
        """
        Convert the server response to an user requested output format on Tables
        :param server_response: Tabular JSON data from server
        :param output_format: format to be converted into
        :param indexing: row & column index consideration in the output
        """
        self.server_response = server_response
        self.output = self._converter(output_format.lower(), indexing=indexing, table_obj=table_obj)

    def _converter(self, fmt: str, indexing: bool = False, table_obj="TableJson") -> list:
        """
        Actual conversion takes place here using Pandas
        :param fmt: format to be converted into
        :param indexing: row index consideration in the output
        :return: list of tables from converted into the requested output format
        """
        dfs = []
        for table in self.server_response.get("Tables", []):
            tmp = {int(k): v for k, v in table[table_obj].items()}
            # To convert column indices to int to maintain the table order with more than 9 columns
            cols = [str(x) for x in sorted([int(x) for x in tmp[0]])] if tmp else None
            # To convert row indices to int and maintain the table order with more than 9 rows
            tmp = collections.OrderedDict(sorted(tmp.items()))
            dfs.append(pd.DataFrame.from_dict(tmp, orient="index", columns=cols))

        if fmt in ("df", "dataframe"):
            return dfs
        elif fmt == "dict":
            return [df.to_dict() for df in dfs]
        elif fmt == "csv":
            save_folder = tempfile.mkdtemp()
            output_location = []
            for tbl_n, df in enumerate(dfs):
                csv_name = os.path.join(save_folder, f"_table_{tbl_n+1}.csv")
                df.to_csv(csv_name, index=indexing, header=indexing)
                output_location.append(csv_name)
            return output_location
        elif fmt in ("xlsx", "excel"):
            output_excel_location = os.path.join(tempfile.mkdtemp(), f"_tables_{len(dfs)}.xlsx")
            if len(dfs) >= 10:
                warnings.warn(f"There are {dfs} tables extracted. Consider to change the output_format to 'csv' instead")
            with pd.ExcelWriter(output_excel_location) as writer:
                for n, df in enumerate(dfs):
                    df.to_excel(writer, f'table_{n+1}', index=indexing, header=indexing)
                writer.save()
            return [output_excel_location]
        elif fmt == "json":
            return [df.to_json() for df in dfs]
        else:
            warn_msg = f"Supported output formats {self.FORMATS} only. Assigned to default: {self.DEFAULT}"
            warnings.warn(warn_msg)
            return dfs


class MakeCorrections:
    def __init__(self, et_resp: dict = None, dataframes: List[pd.DataFrame] = None):
        """
        To apply post processing techniques on the output
        :param et_resp: ExtractTable response
        :param dataframes: user preferred dataframe(s).
            Default assumes all dataframes from the extracttable response, `et_resp`.
            If both `et_resp` and `dataframes` are provided, the later is considered for the processing
        """
        if et_resp:
            self.dataframes = ConvertTo(server_response=et_resp).output

        if not et_resp:
            try:
                self.dataframes = self.__isacceptable__(dataframes)
            except ValueError:
                raise ValueError("Either ExtractTable response or your preferred list of pandas dataframes is required")

    @staticmethod
    def __isacceptable__(dfs) -> List[pd.DataFrame]:
        """Validate the `dataframes` param"""
        if type(dfs) is list:
            if all([type(df) is pd.DataFrame for df in dfs]):
                return dfs
        elif type(dfs) is pd.DataFrame:
            return [dfs]
        raise ValueError("Dataframes should be list of dataframes or a dataframe")

    def split_merged_rows(self) -> List[pd.DataFrame]:
        """
        To split the merged rows into possible multiple rows
        :return: reformatted list of dataframes
        """
        for df_idx, each_df in enumerate(self.dataframes):
            reformat = []
            for row in each_df.to_numpy():
                row = list(row)

                # looks like line separator is " "
                seperators = [col.strip().count(" ") for col in row]
                # Statistical mode to assume the number of rows merged
                mode_ = mode(seperators)

                if mode_:
                    # split the merged rows inside the col
                    tmp = [col.strip().split(' ', mode_) for col in row]
                    for idx in range(len(tmp[0])):
                        tmp_ = []
                        for x in range(len(tmp)):
                            try:
                                val = tmp[x][idx]
                            except IndexError:
                                val = ""
                            tmp_.append(val)
                        reformat.append(tmp_)
                else:
                    reformat.append(row)

            self.dataframes[df_idx] = pd.DataFrame(reformat)

        return self.dataframes

    def split_merged_columns(self, columns_idx: List[int] = None, force_split: bool = False) -> List[pd.DataFrame]:
        """
        To split the merged columns into possible multiple columns
        :param columns_idx: user preferred columns indices.
                Default loops through all columns to find numeric or decimal columns
        :param force_split: To force split through the columns
        :return: reformatted list of dataframes
        """
        # TODO: Should we consider delimiter_pattern for the split?
        for df_idx, df in enumerate(self.dataframes):
            if not columns_idx:
                columns_idx = df.columns

            columns_idx = [str(x) for x in columns_idx]
            reformat = []
            for col_idx in columns_idx:
                tmp = df[col_idx].str.split(expand=True)

                if not any([not any(tmp.isna().any()), force_split]) or tmp.shape[-1] == 1:
                    reformat.append(df[col_idx].tolist())
                    # If user wanted force_split or the split columns have all cell values
                    # then proceed next
                else:
                    reformat.extend([tmp[each].tolist() for each in tmp.columns])

            self.dataframes[df_idx] = pd.DataFrame(reformat).T

        return self.dataframes

    def fix_decimal_format(self, columns_idx: List[int] = None, decimal_separator: str = ".", thousands_separator: str = ",", decimal_position: int = 2) -> List[pd.DataFrame]:
        """
        To fix decimal and thousands separator values. Often commas as detected as period
        :param columns_idx: user preferred columns indices.
                Default loops through all columns to find numeric or decimal columns
        :param decimal_separator: preferred decimal separator
        :param thousands_separator: preferred thousands separator
        :param decimal_position: preferred decimal position
        :return: corrected list of dataframes
        """
        # TODO: Should we consider only bad confidence values?
        reg_ = f"[{decimal_separator}{thousands_separator}]"
        if decimal_position > 0:
            thou_regex = reg_ + '(?=.*' + reg_ + ')'
        else:
            thou_regex = reg_
        decimal_position = int(decimal_position)

        for df_idx, df in enumerate(self.dataframes):
            if not columns_idx:
                columns_idx = df.columns
            columns_idx = [str(x) for x in columns_idx]

            for col_idx in columns_idx:
                digits = df[col_idx].str.count(pat=r'\d').sum()
                chars = df[col_idx].str.count(pat=r'[\w]').sum()

                if digits/chars < 0.75:
                    # To infer a numeric or float column
                    # Check if the column contains more digits or characters
                    continue

                df[col_idx] = df[col_idx].str.strip()
                df[col_idx].replace(regex={r'%s' % thou_regex: thousands_separator}, inplace=True)

                # To correct decimal position
                if not decimal_position > 0:
                    continue

                for i, _ in enumerate(df[col_idx]):
                    if not len(df[col_idx][i]) > decimal_position:
                        # length of atleast decimal_position
                        continue
                    elif df[col_idx][i][-(decimal_position+1)] == decimal_separator:
                        # nothing to do if decimal separator already in place
                        continue

                    # If decimal position is a not alphanumeric
                    if re.search(r'\W+', df[col_idx][i][-(decimal_position+1)]):
                        digits = len(re.findall(r'\d', df[col_idx][i]))
                        if digits/len(df[col_idx][i]) >= 0.5:
                            df[col_idx][i] = df[col_idx][i][:-(decimal_position+1)] + decimal_separator + df[col_idx][i][-decimal_position:]

            self.dataframes[df_idx] = df
        return self.dataframes

    def fix_date_format(self, columns_idx: List[int] = None, delimiter: str = "/"):
        """
        To fix date formats of the column
        Eg: 12|1212020 as 12/12/2020
        :param columns_idx: user preferred columns indices.
                Default loops through all columns to find Date Columns
        :param delimiter: "/" or "-" whatelse you prefer
        :return: correted list of dataframes
        """
        date_regex = r'(\d{2}(\d{2})?)(\W)(\d{2}|[A-Za-z]{3,9})(\W)(\d{2}(\d{2})?)\b'
        for df_idx, df in enumerate(self.dataframes):
            if not columns_idx:
                columns_idx = df.columns
            columns_idx = [str(x) for x in columns_idx]

            for col_idx in columns_idx:
                dates = df[col_idx].str.count(pat=date_regex).sum()

                if not (dates >= len(df) * 0.75):
                    # To infer a date column
                    # Check if the column contains digits and non-alpha character greater than column length
                    continue

                df[col_idx] = df[col_idx].str.strip()
                df[col_idx].replace(regex={date_regex: r'\1%s\4%s\6' % (delimiter, delimiter)}, inplace=True)

            self.dataframes[df_idx] = df

        return self.dataframes



"""
To validate the files at Client End; Fail Faster
"""
import typing as ty
import os
import shutil
import tempfile

import requests
import PyPDF2

from ..exceptions import *


class CheckFile:
    __SUPPORTED_EXTENSIONS__: tuple = tuple({'.pdf', '.jpeg', '.jpg', '.png'})
    __THRESHOLD_SIZE__: int = 4      # Megabytes

    def __init__(self, filepath: ty.Union[os.PathLike, str]):
        self.filepath = filepath
        self.type_error()
        self.size_error()

    def type_error(self) -> ty.Union[Exception, None]:
        """To check file extension"""
        if self.filepath.lower().endswith(self.__SUPPORTED_EXTENSIONS__):
            return
        raise ClientFileTypeError(Message=f"Allowed file types are {self.__SUPPORTED_EXTENSIONS__}")

    def size_error(self) -> ty.Union[Exception, None]:
        # 1027 to create some buffer
        if os.stat(self.filepath).st_size <= self.__THRESHOLD_SIZE__*1027*1027:
            return
        raise ClientFileSizeError(Message=f"File Size greater than the threshold {self.__THRESHOLD_SIZE__} Mb.")


class PrepareInput:
    """
    Handle PDF work
    """
    def __enter__(self):
        return self

    def __init__(self, filepath: ty.Union[os.PathLike, str], pages: str):
        self.filepath = filepath
        self.temp_dir = tempfile.mkdtemp()
        if self.filepath.startswith(("http://", "https://")):
            self.filepath = self.download_file(self.filepath)
        self.pages = pages
        # Save time by using the real file,
        # if "all" pages or an image file
        if pages == "all" or not self.filepath.lower().endswith(".pdf"):
            pass
        else:
            print("[Info]: Aggregating user defined pages..", self.pages)
            gather_pages = self._get_pages(self.filepath, pages)
            self.filepath = self.pdf_separator(gather_pages)
        CheckFile(self.filepath)

    def pdf_separator(self, gather_pages: set):
        """PDF Splitter"""
        merged_pdf = os.path.join(self.temp_dir, str(self.pages) + os.path.basename(self.filepath))
        with open(merged_pdf, 'wb') as out_file:
            pdf_reader = PyPDF2.PdfFileReader(self.filepath)
            pdf_writer = PyPDF2.PdfFileWriter()
            for page in gather_pages:
                try:
                    pdf_writer.addPage(pdf_reader.getPage(page-1))
                except IndexError:
                    raise EOFError(f"File has only {pdf_reader.numPages} pages, but asked for {self.pages}")
            pdf_writer.write(out_file)
        return merged_pdf

    @staticmethod
    def _get_pages(filepath: os.PathLike, pages: str) -> set:
        # Credits to camelot-py library - customized
        """Converts pages string to list of ints.

        Parameters
        ----------
        filepath : Pathlike
            Filepath or URL of the PDF file.
        pages : str, optional (default: '1')
            Comma-separated page numbers.
            Example: '1,3,4' or '1,4-end' or 'all'.

        Returns
        -------
        List of int page numbers.

        """
        page_numbers = []
        pages_needed = []

        if pages == "1":
            page_numbers.append({"start": 1, "end": 1})
        else:
            with open(filepath, "rb") as file_obj:
                infile = PyPDF2.PdfFileReader(file_obj, strict=False)
                if pages == "all":
                    page_numbers.append({"start": 1, "end": infile.getNumPages()})
                else:
                    for r in pages.split(","):
                        if "-" in r:
                            a, b = r.split("-")
                            if b == "end":
                                b = infile.getNumPages()
                            page_numbers.append({"start": int(a), "end": int(b)})
                        else:
                            page_numbers.append({"start": int(r), "end": int(r)})

        for p in page_numbers:
            pages_needed.extend(range(p["start"], p["end"] + 1))

        return set(pages_needed)

    def download_file(self, url: str):
        """
        Download file to local
        :param url: PDF file path
        :return: downloaded file local filepath
        """
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            _, r_ext = r.headers['Content-Type'].rsplit('/', 1)
            fname, f_ext = os.path.basename(url).rsplit('.', 1)
            ext = r_ext if r_ext else f_ext
            ext = ext.lower()
            # TODO use filetype lib to find extension
            tmp_fname = os.path.join(self.temp_dir, f"{fname}.{ext}")
            with open(tmp_fname, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:   # filter out keep-alive new chunks
                        f.write(chunk)
                        f.flush()
        return tmp_fname

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Delete the temporary directory created for an instance"""
        if self.temp_dir:
            shutil.rmtree(self.temp_dir)



"""
Create custom exceptions and warnings here
"""


class ServiceError(Exception):
    """Raise a Service Side error based on the response received"""
    def __init__(self, dict_resp: dict = None):
        print("-#- "*10)
        print("-"*9 + "Error Server Response" + "-"*9)
        self.resp = dict_resp
        for k, v in self.resp.items():
            print(f"{k}: {v}")
            self.__setattr__(k, v)
        print("-#- "*10)

    def __str__(self) -> str:
        _msg_ = ''
        if self.resp.get('Message', ''):
            _msg_ = self.resp.get('Message', '')
        else:
            _msg_ = self.resp.get('message', '')
        return str(_msg_)


class ClientFileSizeError(Exception):
    """Raise a Client Side error based on the file to be processed"""
    def __init__(self, **kwargs):
        self.resp = kwargs

    def __str__(self) -> str:
        _msg_ = ''
        if self.resp.get('Message', ''):
            _msg_ = self.resp.get('Message', '')
        else:
            _msg_ = self.resp.get('message', '')
        return str(_msg_)


class ClientFileTypeError(Exception):
    """Raise a Client Side error based on the file to be processed"""
    def __init__(self, **kwargs):
        self.resp = kwargs

    def __str__(self) -> str:
        _msg_ = ''
        if self.resp.get('Message', ''):
            _msg_ = self.resp.get('Message', '')
        else:
            _msg_ = self.resp.get('message', '')
        return str(_msg_)



"""
Configure all Server request/response objects here
"""


class HOST:
    """API Endpoints of ExtractTable.com"""
    VALIDATOR = 'validator.extracttable.com'
    TRIGGER = 'trigger.extracttable.com'
    RESULT = 'getresult.extracttable.com'
    BIGFILE = 'bigfile.extracttable.com'


class JobStatus:
    """Job Status responses recieved from Server. Declared here to maintain consistency"""
    SUCCESS = 'Success'
    FAILED = 'Failed'
    PROCESSING = 'Processing'
    INCOMPLETE = 'Incomplete'
